package eu.jonahbauer.raytracing.scene;

import eu.jonahbauer.raytracing.math.AABB;
import eu.jonahbauer.raytracing.math.Range;
import eu.jonahbauer.raytracing.math.Ray;
import eu.jonahbauer.raytracing.math.Vec3;
import eu.jonahbauer.raytracing.scene.transform.RotateY;
import eu.jonahbauer.raytracing.scene.transform.Translate;
import org.jetbrains.annotations.NotNull;

import java.util.Optional;

public interface Hittable {

    /**
     * @see #hit(Ray, Range)
     */
    default @NotNull Optional<HitResult> hit(@NotNull Ray ray) {
        return hit(ray, new Range(0.001, Double.POSITIVE_INFINITY));
    }

    /**
     * Tests whether the {@code ray} intersects {@code this} hittable.
     * <p>
     * The second parameter {@code range} allows the implementation to skip unnecessary calculations if it can
     * determine that a hit (if any) will fall outside the valid range of {@code t}s. The returned hit may still be
     * outside the valid range and has to be checked by the caller.
     * @param ray a ray
     * @param range the range of valid {@code t}s
     * @return the result of the hit test, containing (among others) the value {@code t} such that {@code ray.at(t)} is
     *         a point on {@code this} hittable
     */
    @NotNull Optional<HitResult> hit(@NotNull Ray ray, @NotNull Range range);

    /**
     * {@return the axis-aligned bounding box of this hittable}
     */
    @NotNull AABB getBoundingBox();

    default @NotNull Hittable translate(@NotNull Vec3 offset) {
        return Translate.create(this, offset);
    }

    default @NotNull Hittable rotateY(double angle) {
        return RotateY.create(this, angle);
    }
}
